#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/trace.py
# DESCRIPTION:    Trace/audit for class instances
# CREATED:        5.6.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""firebird-base - Trace/audit for class instances
"""

from __future__ import annotations
from typing import Any, Type, Hashable, List, Dict, Callable
import os
from inspect import signature, Signature
from dataclasses import dataclass, field
from enum import IntFlag, auto
from functools import wraps
from time import monotonic
from decimal import Decimal
from .types import Distinct, DEFAULT, UNLIMITED
from .collections import Registry
from .strconv import convert_from_str
from .logging import LogLevel, FBLoggerAdapter, get_logger

class TraceFlag(IntFlag):
    "`LoggingManager` trace/audit flags"
    DISABLED = 0
    ACTIVE = auto()
    BEFORE = auto()
    AFTER = auto()
    FAIL = auto()

@dataclass
class TracedItem(Distinct):
    "Class method trace specification"
    method: str
    decorator: Callable
    args: List = field(default_factory=list)
    kwargs: Dict = field(default_factory=dict)
    def get_key(self) -> Hashable:
        return self.method

@dataclass
class TracedClass(Distinct):
    "Traced class registry entry"
    cls: Type
    traced: Registry = field(default_factory=Registry)
    def get_key(self) -> Hashable:
        return self.cls

_traced: Registry = Registry()

class TracedMeta(type):
    "Metaclass that instruments instances on creation."
    def __call__(cls: Type, *args, **kwargs):
        return trace_object(super().__call__(*args, **kwargs), strict=True)

class TracedMixin(metaclass=TracedMeta):
    """Mixin class that automatically registers descendants for trace and instruments
instances on creation.
"""
    def __init_subclass__(cls: Type, /, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        trace_manager.register(cls)


class traced:
    """Base decorator for logging of callables, suitable for trace/audit.

It's not applied on decorated function/method if `FBASE_TRACE` environment variable is
set to False, or if `FBASE_TRACE` is not defined and `__debug__` is False (optimized
Python code).

Both positional and keyword arguments of decorated callable are available by name for
f-string type message interpolation as `dict` passed to logger as positional argument.
"""
    def __init__(self, *, agent: Any=DEFAULT, context: Any=DEFAULT, topic: str='trace',
                 msg_before: str=DEFAULT, msg_after: str=DEFAULT, msg_failed: str=DEFAULT,
                 flags: TraceFlag=TraceFlag(0), level: LogLevel=LogLevel.DEBUG,
                 max_param_length: int=UNLIMITED, extra: Dict=None,
                 callback: Callable[[Any], bool]=None, has_result: bool=DEFAULT,
                 with_args: bool=True):
        """
        Arguments:
            agent: Agent identification
            context: Context identification
            topic: Trace/audit logging topic
            msg_before: Trace/audit message logged before decorated function
            msg_after: Trace/audit message logged after decorated function
            msg_failed: Trace/audit message logged when decorated function raises an exception
            flags: Trace flags override
            level: Logging level for trace/audit messages
            max_param_length: Max. length of parameters (longer will be trimmed)
            extra: Extra data for `LogRecord`
        """
        self.msg_before: str = msg_before
        self.msg_after: str = msg_after
        self.msg_failed: str = msg_failed
        self.agent: Any = agent
        self.context: Any = context
        self.topic: str = topic
        self.flags: TraceFlag = flags
        self.level: LogLevel = level
        self.max_len: int = max_param_length
        self.extra: Dict = extra
        self.callback: Callable[[Any], bool] = self.__callback if callback is None else callback
        self.has_result: bool = has_result
        self.with_args: bool = with_args
    def __callback(self, agent: Any) -> bool:
        "Default callback, does nothing"
        return True
    def set_before_msg(self, fn: Callable, sig: Signature) -> None:
        "Sets the DEFAULT before message f-string template."
        if self.with_args:
            self.msg_before = f">>> {fn.__name__}({', '.join(f'{{{x}=}}' for x in sig.parameters if x != 'self')})"
        else:
            self.msg_before = f">>> {fn.__name__}"
    def set_after_msg(self, fn: Callable, sig: Signature) -> None:
        "Sets the DEFAULT after message f-string template."
        self.msg_after = f"<<< {fn.__name__}[{{_etime_}}] Result: {{_result_}}" \
            if self.has_result else f"<<< {fn.__name__}[{{_etime_}}]"
    def set_fail_msg(self, fn: Callable, sig: Signature) -> None:
        "Sets the DEFAULT fail message f-string template."
        self.msg_failed = f"<-- {fn.__name__}[{{_etime_}}] {{_exc_}}"
    def log_before(self, logger: FBLoggerAdapter, params: Dict) -> None:
        "Executed before decorated callable"
        logger.log(self.level, self.msg_before, params, stacklevel=2)
    def log_after(self, logger: FBLoggerAdapter, params: Dict) -> None:
        "Executed after decorated callable"
        logger.log(self.level, self.msg_after, params, stacklevel=2)
    def log_failed(self, logger: FBLoggerAdapter, params: Dict) -> None:
        "Executed when decorated callable raises an exception"
        logger.log(self.level, self.msg_failed, params, stacklevel=2)
    def __call__(self, fn: Callable):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            flags = trace_manager.trace | self.flags
            if enabled := ((TraceFlag.ACTIVE in flags) and int(flags) > 1):
                params = {}
                bound = sig.bind_partial(*args, **kwargs)
                # If it's not a bound method, look for 'self'
                log = get_logger(bound.arguments.get('self', 'function') if self.agent is None
                                 else self.agent, self.context, self.topic)
                if enabled := (log.isEnabledFor(self.level) and self.callback(self.agent)):
                    if self.with_args:
                        bound.apply_defaults()
                        params.update(bound.arguments)
                        if self.max_len is not UNLIMITED:
                            for k, v in params.items():
                                s = str(v)
                                if (i := len(s)) > self.max_len:
                                    params[k] = f'{s[:self.max_len]}..[{i - self.max_len}]'
                    if self.extra is not None:
                        params.update(self.extra)
                    params['_fname_'] = fn.__name__
                    params['_result_'] = None
                    #
                    if TraceFlag.BEFORE in flags:
                        self.log_before(log, params)
            try:
                result = None
                start = monotonic()
                result = fn(*args, **kwargs)
            except Exception as exc:
                if enabled and TraceFlag.FAIL | TraceFlag.ACTIVE in flags:
                    e = str(Decimal(monotonic() - start))
                    params['_etime_'] = e[:e.find('.')+6]
                    params['_exc_'] = f'{exc.__class__.__qualname__}: {exc}'
                    self.log_failed(log, params)
                raise
            else:
                if enabled and TraceFlag.AFTER | TraceFlag.ACTIVE in flags:
                    e = str(Decimal(monotonic() - start))
                    params['_etime_'] = e[:e.find('.')+6]
                    if self.has_result:
                        params['_result_'] = result
                        if self.max_len is not UNLIMITED:
                            s = str(result)
                            if (i := len(s)) > self.max_len:
                                params['_result_'] = f'{s[:self.max_len]}..[{i - self.max_len}]'
                    self.log_after(log, params)
            return result

        if (trace := os.getenv('FBASE_TRACE')) is not None:
            if not convert_from_str(bool, trace):
                return fn
        elif not __debug__:
            return fn
        if self.agent is DEFAULT:
            self.agent = getattr(fn, '__self__', None)
        sig = signature(fn)
        if self.has_result is DEFAULT:
            self.has_result = sig.return_annotation != 'None'
        if self.msg_before is DEFAULT:
            self.set_before_msg(fn, sig)
        if self.msg_after is DEFAULT:
            self.set_after_msg(fn, sig)
        if self.msg_failed is DEFAULT:
            self.set_fail_msg(fn, sig)
        return wrapper

class TraceManager:
    """Trace manager."""
    def __init__(self):
        self._traced: Registry = Registry()
        self._flags: TraceFlag = TraceFlag.DISABLED
        if convert_from_str(bool, os.getenv('FBASE_TRACE', str(__debug__))):
            self.trace |= TraceFlag.ACTIVE
        if convert_from_str(bool, os.getenv('FBASE_TRACE_BEFORE', 'no')): # pragma: no cover
            self.trace |= TraceFlag.BEFORE
        if convert_from_str(bool, os.getenv('FBASE_TRACE_AFTER', 'no')): # pragma: no cover
            self.trace |= TraceFlag.AFTER
        if convert_from_str(bool, os.getenv('FBASE_TRACE_FAIL', 'yes')):
            self.trace |= TraceFlag.FAIL
    def is_registered(self, cls: Type) -> bool:
        "Return True if class is registered"
        return cls in self._traced
    def clear(self) -> None:
        "Removes all trace specifications."
        for cls in self._traced:
            cls.traced.clear()
    def register(self, cls: Type) -> None:
        "Register class for trace. Does nothing if class is already registered."
        if cls not in self._traced:
            self._traced.store(TracedClass(cls))
    def add_trace(self, cls: Type, method: str, decorator: Callable=traced, / , *args, **kwargs) -> None:
        "Add/update trace specification for class method."
        self._traced[cls].traced.update(TracedItem(method, decorator, args, kwargs))
    def remove_trace(self, cls: Type, method: str) -> None:
        "Remove trace specification for class method."
        del self._traced[cls].traced[method]
    def trace_object(self, obj: Any, *, strict: bool=False) -> Any:
        """Instruments object's methods with decorator according to trace configuration.

Arguments:
    strict: Determines the response if the object class is not registered for trace.
            Raises exception when True, or return the instance as is when False [default].

Returns:
    Decorated instance.

Raises:
    TypeError: When object class is not registered and `strict` is True.
"""
        if (trace := os.getenv('FBASE_TRACE')) is not None:
            if not convert_from_str(bool, trace):
                return obj
        elif not __debug__:
            return obj
        entry: TracedClass = self._traced.get(obj.__class__)
        if entry is None:
            if strict:
                raise TypeError(f"Class '{obj.__class__.__name__}' not registered for trace!")
            else:
                return obj
        for item in entry.traced:
            setattr(obj, item.method, item.decorator(*item.args, **item.kwargs)(getattr(obj, item.method)))
        return obj
    @property
    def trace(self) -> TraceFlag:
        "Trace flags"
        return self._flags
    @trace.setter
    def trace(self, value: TraceFlag) -> None:
        self._flags = value if isinstance(value, TraceFlag) else TraceFlag(value)

#: Trace manager
trace_manager: TraceManager = TraceManager()

add_trace = trace_manager.add_trace
remove_trace = trace_manager.remove_trace
trace_object = trace_manager.trace_object




#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/hooks.py
# DESCRIPTION:    Hook manager
# CREATED:        14.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Firebird Base - Hook manager

This module provides a general framework for callbacks and "hookable" events.
"""

from __future__ import annotations
from typing import Union, Any, Type, Dict, List, Set, Callable, cast
from enum import Enum, Flag, auto
from weakref import WeakKeyDictionary
from dataclasses import dataclass, field
from .types import Distinct, ANY, Singleton
from .collections import Registry

@dataclass(order=True, frozen=True)
class Hook(Distinct):
    "Hook registration info"
    event: Any
    cls: type = ANY
    instance: Any = ANY
    callbacks: List = field(default_factory=list)
    def get_key(self) -> Any:
        "Returns hook key"
        return (self.event, self.cls, self.instance)

class HookFlag(Flag):
    "Internally used flags"
    NONE = 0
    INSTANCE = auto()
    CLASS = auto()
    NAME = auto()
    ANY_EVENT = auto()

class HookManager(Singleton):
    """Hook manager."""
    def __init__(self):
        self.obj_map: WeakKeyDictionary = WeakKeyDictionary()
        self.hookables: Dict[Type, Set[Any]] = {}
        self.hooks: Registry = Registry()
        self.flags: HookFlag = HookFlag.NONE
    def _update_flags(self, event: Any, cls: Any, obj: Any) -> None:
        if event is ANY:
            self.flags |= HookFlag.ANY_EVENT
        if cls is not ANY:
            self.flags |= HookFlag.CLASS
        if obj is not ANY:
            if isinstance(obj, str):
                self.flags |= HookFlag.NAME
            else:
                self.flags |= HookFlag.INSTANCE
    def register_class(self, cls: Type, events: Union[Type[Enum], Set]=None) -> None:
        """Register hookable class.

Arguments:
    cls: Class that supports hooks.
    events: Supported events.

Events could be specified using an `~enum.Enum` type or set of event identificators.
When Enum is used (recommended), all enum values are registered as hookable events.
"""
        if isinstance(events, type) and issubclass(events, Enum):
            events = set(events.__members__.values())
        self.hookables[cls] = events
    def register_name(self, instance: Any, name: str) -> None:
        """Associate name with hookable instance.

Arguments:
    instance: Instance of registered hookable class.
    name:     Unique name assigned to instance.
"""
        if not isinstance(instance, tuple(self.hookables.keys())):
            raise TypeError("The instance is not of hookable type")
        self.obj_map[instance] = name
    def add_hook(self, event: Any, source: Any, callback: Callable) -> None:
        """Add new hook.

Arguments:
    event:    Event identificator.
    source:   Hookable class or instance, or instance name.
    callback: Callback function.

Important:
    The signature of `callback` must conform to requirements for particular hookable event.

Raises:
    TypeError: When `subject` is not registered as hookable.
    ValueError: When `event` is not supported by specified `subject`.
"""
        cls = obj = ANY
        if isinstance(source, type):
            if source in self.hookables:
                cls = source
                if event is not ANY:
                    found = False
                    for cls_ in (c for c in self.hookables if issubclass(cls, c)):
                        if event in self.hookables[cls_]:
                            found = True
                            break
                    if not found:
                        raise ValueError(f"Event '{event}' is not supported by '{cls.__name__}'")
            else:
                raise TypeError("The type is not registered as hookable")
        elif isinstance(source, tuple(self.hookables)):
            obj = source
            if event is not ANY:
                found = False
                for cls_ in (c for c in self.hookables if isinstance(obj, c)):
                    if event in self.hookables[cls_]:
                        found = True
                        break
                if not found:
                    raise ValueError(f"Event '{event}' is not supported by '{obj.__class__.__name__}'")
        elif isinstance(source, str):
            obj = source
        else:
            raise TypeError("Subject must be hookable class or instance, or name")
        self._update_flags(event, cls, obj)
        key = (event, cls, obj)
        hook: Hook = self.hooks[key] if key in self.hooks else self.hooks.store(Hook(*key))
        hook.callbacks.append(callback)
    def remove_hook(self, event: Any, source: Any, callback: Callable) -> None:
        """Remove hook callback installed by `add_hook()`.

Arguments:
    event:    Event identificator.
    source:   Hookable class or instance.
    callback: Callback function.

Important:
    For successful removal, the argument values must be exactly the same as used in
    `add_hook()` call.

    The method does nothing if described hook is not installed.
"""
        cls = obj = ANY
        if isinstance(source, type):
            cls = source
        else:
            obj = source
        key = (event, cls, obj)
        hook: Hook = self.hooks.get(key)
        if hook is not None:
            hook.callbacks.remove(callback)
            if not hook.callbacks:
                self.hooks.remove(hook)
                self.flags = HookFlag.NONE
                for h in self.hooks:
                    self._update_flags(h.event, h.cls, h.instance)
    def remove_all_hooks(self) -> None:
        "Removes all installed hooks."
        self.hooks.clear()
        self.flags = HookFlag.NONE
    def reset(self) -> None:
        """Removes all installed hooks and unregisters all hookable classes and instances."""
        self.remove_all_hooks()
        self.hookables.clear()
        self.obj_map.clear()
    def get_callbacks(self, event: Any, source: Any) -> List:
        """Returns list of all callbacks installed for specified event and hookable subject.

Arguments:
    event:  Event identificator.
    source: Hookable class or instance, or name.
"""
        result = []
        if isinstance(source, type):
            if HookFlag.CLASS in self.flags:
                if (hook := self.hooks.get((event, source, ANY))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
                if HookFlag.ANY_EVENT in self.flags and (hook := self.hooks.get((ANY, source, ANY))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
        elif isinstance(source, str):
            if HookFlag.NAME in self.flags:
                if (hook := self.hooks.get((event, ANY, source))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
                if HookFlag.ANY_EVENT in self.flags and (hook := self.hooks.get((ANY, ANY, source))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
        else:
            if HookFlag.INSTANCE in self.flags:
                if (hook := self.hooks.get((event, ANY, source))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
                if HookFlag.ANY_EVENT in self.flags and (hook := self.hooks.get((ANY, ANY, source))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
            if HookFlag.NAME in self.flags and (name := self.obj_map.get(source)) is not None:
                if (hook := self.hooks.get((event, ANY, name))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
                if HookFlag.ANY_EVENT in self.flags and (hook := self.hooks.get((ANY, ANY, name))) is not None:
                    result.extend(cast(Hook, hook).callbacks)
            if HookFlag.CLASS in self.flags:
                for cls in (c for c in self.hookables if isinstance(source, c)):
                    if (hook := self.hooks.get((event, cls, ANY))) is not None:
                        result.extend(cast(Hook, hook).callbacks)
                    if HookFlag.ANY_EVENT in self.flags and (hook := self.hooks.get((ANY, cls, ANY))) is not None:
                        result.extend(cast(Hook, hook).callbacks)
        return result

#: Hook manager
hook_manager: HookManager = HookManager()

register_class = hook_manager.register_class
register_name = hook_manager.register_name
add_hook = hook_manager.add_hook
get_callbacks = hook_manager.get_callbacks



#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/protobuf.py
# DESCRIPTION:    Registry for Google Protocol Buffer messages and enums
# CREATED:        27.12.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Firebird Base - Registry for Google Protocol Buffer messages and enums
"""

from __future__ import annotations
from typing import Any, Callable, cast
from dataclasses import dataclass
from pkg_resources import iter_entry_points
from google.protobuf.descriptor import EnumDescriptor
from .types import Distinct
from .collections import Registry

# Classes

@dataclass(eq=True, order=True, frozen=True)
class ProtoMessageType(Distinct):
    """Google protobuf message type"""
    name: str
    constructor: Callable
    def get_key(self) -> Any:
        """Returns `name`."""
        return self.name

@dataclass(eq=True, order=True, frozen=True)
class ProtoEnumType(Distinct):
    """Google protobuf enum type
"""
    descriptor: EnumDescriptor
    def get_key(self) -> Any:
        """Returns `name`."""
        return self.name
    def __getattr__(self, name):
        """Returns the value corresponding to the given enum name."""
        if name in self.descriptor.values_by_name:
            return self.descriptor.values_by_name[name].number
        raise AttributeError(f"Enum {self.name} has no value with name '{name}'")
    def keys(self):
        """Return a list of the string names in the enum.

        These are returned in the order they were defined in the .proto file.
        """
        return [value_descriptor.name for value_descriptor in self.descriptor.values]
    def values(self):
        """Return a list of the integer values in the enum.

        These are returned in the order they were defined in the .proto file.
        """
        return [value_descriptor.number for value_descriptor in self.descriptor.values]
    def items(self):
        """Return a list of the (name, value) pairs of the enum.

        These are returned in the order they were defined in the .proto file.
        """
        return [(value_descriptor.name, value_descriptor.number)
                for value_descriptor in self.descriptor.values]
    def get_value_name(self, number: int) -> str:
        """Returns a string containing the name of an enum value.

        Raises:
            KeyError: If there is no value for specified name.
        """
        if number in self.descriptor.values_by_number:
            return self.descriptor.values_by_number[number].name
        raise KeyError(f"Enum {self.name} has no name defined for value {number}")
    @property
    def name(self) -> str:
        "Returns full enum type name"
        return self.descriptor.full_name

_msgreg: Registry = Registry()
_enumreg: Registry = Registry()

def create_message(name: str) -> Any:
    """Returns new protobuf message instance.

Arguments:
    name: Fully qualified protobuf message name.

Raises:
    KeyError: When message type is not registered.
"""
    if (msg := _msgreg.get(name)) is None:
        raise KeyError(f"Unregistered protobuf message '{name}'")
    return cast(ProtoMessageType, msg).constructor()

def is_msg_registered(name: str) -> bool:
    """Returns True if specified `name` refers to registered protobuf message type."""
    return name in _msgreg

def is_enum_registered(name: str) -> bool:
    """Returns True if specified `name` refers to registered protobuf enum type."""
    return name in _enumreg

def get_enum_type(name: str) -> ProtoEnumType:
    """Returns wrapper instance for protobuf enum type with specified `name`.

Raises:
    KeyError: When enum type is not registered.
"""
    if (e := _enumreg.get(name)) is None:
        raise KeyError(f"Unregistered protobuf enum type '{name}'")
    return e

def get_enum_field_type(msg, field_name: str) -> str:
    """Returns name of enum type for message enum field.

Raises:
    KeyError: When message does not have specified field.
"""
    if (fdesc := msg.DESCRIPTOR.fields_by_name.get(field_name)) is None:
        raise KeyError(f"Message does not have field '{field_name}'")
    return fdesc.enum_type.full_name

def get_enum_value_name(enum_type_name: str, value: Any) -> str:
    """Returns name for the enum value."""
    return get_enum_type(enum_type_name).get_value_name(value)


def register_decriptor(file_descriptor) -> None:
    "Registers enums and messages defined by protobuf file DESCRIPTOR."
    for msg_desc in file_descriptor.message_types_by_name.values():
        _msgreg.store(ProtoMessageType(msg_desc.full_name, msg_desc._concrete_class))
    for enum_desc in file_descriptor.enum_types_by_name.values():
        _enumreg.store(ProtoEnumType(enum_desc))

def load_registered(group: str) -> None: # pragma: no cover
    """Load registered protobuf packages.

Protobuf packages must register the pb2-file DESCRIPTOR in `entry_point` section of
`setup.cfg` file.

Arguments:
    group: Entry-point group name.

Example:
    ::

       # setup.cfg:

       [options.entry_points]
       firebird.base.protobuf =
           firebird.base.lib_a = firebird.base.lib_a_pb2:DESCRIPTOR
           firebird.base.lib_b = firebird.base.lib_b_pb2:DESCRIPTOR
           firebird.base.lib_c = firebird.base.lib_c_pb2:DESCRIPTOR

       # will be loaded with:

       load_registered('firebird.base.protobuf')
"""
    for desc in (entry.load() for entry in iter_entry_points(group)):
        register_decriptor(desc)



#!/usr/bin/python
#coding:utf-8
#
#   PROGRAM/MODULE: firebird-base
#   FILE:           test/test_hooks.py
#   DESCRIPTION:    Unit tests for firebird.base.hooks
#   CREATED:        14.5.2020
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#
#  Copyright (c) Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): Pavel Císař (original code)
#                  ______________________________________.
#
# See LICENSE.TXT for details.

"Firebird Base - Unit tests for firebird.base.hooks."

from __future__ import annotations
from typing import Protocol, List, cast
import unittest
from enum import Enum, auto
from firebird.base.hooks import hook_manager, HookFlag
from firebird.base.types import ANY

class MyEvents(Enum):
    CREATE = auto()
    ACTION = auto()

class with_print(Protocol):
    def print(self, msg: str) -> None:
        ...

class MyHookable:
    def __init__(self, owner: with_print, name: str, *, register: bool=False,
                 use_class: bool=False, use_name: bool=False):
        self.owner = owner
        self.name: str = name
        if register:
            hook_manager.register_name(self, name)
        subj = self
        if use_class:
            subj = MyHookable
        elif use_name:
            subj = name
        for hook in hook_manager.get_callbacks(MyEvents.CREATE, subj):
            try:
                hook(self, MyEvents.CREATE)
            except Exception as e:
                self.owner.print(f"{self.name}.CREATE hook call outcome: ERROR ({e.args[0]})")
            else:
                self.owner.print(f"{self.name}.CREATE hook call outcome: OK")
    def action(self):
        self.owner.print(f"{self.name}.ACTION!")
        for hook in hook_manager.get_callbacks(MyEvents.ACTION, self):
            try:
                hook(self, MyEvents.ACTION)
            except Exception as e:
                self.owner.print(f"{self.name}.ACTION hook call outcome: ERROR ({e.args[0]})")
            else:
                self.owner.print(f"{self.name}.ACTION hook call outcome: OK")

class MySuperHookable(MyHookable):
    def super_action(self):
        self.owner.print(f"{self.name}.SUPER-ACTION!")
        for hook in hook_manager.get_callbacks('super-action', self):
            try:
                hook(self, 'super-action')
            except Exception as e:
                self.owner.print(f"{self.name}.SUPER-ACTION hook call outcome: ERROR ({e.args[0]})")
            else:
                self.owner.print(f"{self.name}.SUPER-ACTION hook call outcome: OK")

class MyHook:
    def __init__(self, owner: with_print, name: str):
        self.owner = owner
        self.name: str = name
    def callback(self, subject: MyHookable, event: MyEvents):
        self.owner.print(f"Hook {self.name} event {event.name if isinstance(event, Enum) else event} called by {subject.name}")
    def err_callback(self, subject: MyHookable, event: MyEvents):
        self.callback(subject, event)
        raise Exception("Error in hook")

def iter_class_properties(cls):
    """Iterator function.

    Args:
        cls (class): Class object.

    Yields:
        `name', 'property` pairs for all properties in class.
"""
    for varname in vars(cls):
        value = getattr(cls, varname)
        if isinstance(value, property):
            yield varname, value

def iter_class_variables(cls):
    """Iterator function.

    Args:
        cls (class): Class object.

    Yields:
        Names of all non-callable attributes in class.
"""
    for varname in vars(cls):
        value = getattr(cls, varname)
        if not (isinstance(value, property) or callable(value)) and not varname.startswith('_'):
            yield varname


class TestHooks(unittest.TestCase):
    """Unit tests for firebird.base.hooks.HookManager"""
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.output: List = []
    def setUp(self) -> None:
        self.output.clear()
        hook_manager.reset()
    def tearDown(self):
        pass
    def print(self, msg: str) -> None:
        self.output.append(msg)
    def test_aaa_hooks(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        self.assertTupleEqual(tuple(hook_manager.hookables.keys()), (MyHookable, ))
        self.assertSetEqual(hook_manager.hookables[MyHookable],
                            set(x for x in cast(Enum, MyEvents).__members__.values()))
        # Optimizations
        self.assertNotIn(HookFlag.CLASS, hook_manager.flags)
        self.assertNotIn(HookFlag.INSTANCE, hook_manager.flags)
        self.assertNotIn(HookFlag.ANY_EVENT, hook_manager.flags)
        self.assertNotIn(HookFlag.NAME, hook_manager.flags)
        # Install hooks
        hook_A: MyHook = MyHook(self, 'Hook-A')
        hook_B: MyHook = MyHook(self, 'Hook-B')
        hook_C: MyHook = MyHook(self, 'Hook-C')
        hook_N: MyHook = MyHook(self, 'Hook-N')
        #
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_A.callback)
        self.assertIn(HookFlag.CLASS, hook_manager.flags)
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_B.err_callback)
        hook_manager.add_hook(MyEvents.ACTION, MyHookable, hook_C.callback)
        hook_manager.add_hook(MyEvents.ACTION, 'Source-A', hook_N.callback)
        self.assertIn(HookFlag.NAME, hook_manager.flags)
        #
        key = (MyEvents.CREATE, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        self.assertIn(hook_B.err_callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.ACTION, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_C.callback, hook_manager.hooks[key].callbacks)
        # Create event sources, emits CREATE
        self.output.clear()
        src_A: MyHookable = MyHookable(self, 'Source-A', register=True)
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: ERROR (Error in hook)'])
        self.output.clear()
        src_B: MyHookable = MyHookable(self, 'Source-B', register=True)
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by Source-B',
                              'Source-B.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by Source-B',
                              'Source-B.CREATE hook call outcome: ERROR (Error in hook)'])
        # Install instance hooks
        hook_manager.add_hook(MyEvents.ACTION, src_A, hook_A.callback)
        self.assertIn(HookFlag.INSTANCE, hook_manager.flags)
        hook_manager.add_hook(MyEvents.ACTION, src_B, hook_B.callback)
        #
        key = (MyEvents.ACTION, ANY, src_A)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.ACTION, ANY, src_B)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_B.callback, hook_manager.hooks[key].callbacks)
        # And action!
        self.output.clear()
        src_A.action()
        self.assertListEqual(self.output,
                             ['Source-A.ACTION!',
                              'Hook Hook-A event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: OK',
                              'Hook Hook-N event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: OK',
                              'Hook Hook-C event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: OK'])
        #
        self.output.clear()
        src_B.action()
        self.assertListEqual(self.output,
                             ['Source-B.ACTION!',
                              'Hook Hook-B event ACTION called by Source-B',
                              'Source-B.ACTION hook call outcome: OK',
                              'Hook Hook-C event ACTION called by Source-B',
                              'Source-B.ACTION hook call outcome: OK'])
        # Optimizations
        self.assertIn(HookFlag.CLASS, hook_manager.flags)
        self.assertIn(HookFlag.INSTANCE, hook_manager.flags)
        self.assertNotIn(HookFlag.ANY_EVENT, hook_manager.flags)
        self.assertIn(HookFlag.NAME, hook_manager.flags)
        # Remove hooks
        hook_manager.remove_hook(MyEvents.CREATE, MyHookable, hook_A.callback)
        key = (MyEvents.CREATE, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertNotIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        hook_manager.remove_hook(MyEvents.CREATE, MyHookable, hook_B.err_callback)
        self.assertFalse(key in hook_manager.hooks)
        #
        hook_manager.remove_hook(MyEvents.ACTION, src_A, hook_A.callback)
        key = (MyEvents.ACTION, ANY, src_A)
        self.assertFalse(key in hook_manager.hooks)
        #
        hook_manager.remove_all_hooks()
        self.assertEqual(len(hook_manager.hooks), 0)
        #
        hook_manager.add_hook(MyEvents.ACTION, MyHookable, hook_C.callback)
        hook_manager.reset()
        self.assertEqual(len(hook_manager.hookables), 0)
        self.assertEqual(len(hook_manager.hooks), 0)
    def test_inherited_hookable(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        self.assertTupleEqual(tuple(hook_manager.hookables.keys()), (MyHookable, ))
        self.assertSetEqual(hook_manager.hookables[MyHookable],
                            set(x for x in cast(Enum, MyEvents).__members__.values()))
        # Install hooks
        hook_A: MyHook = MyHook(self, 'Hook-A')
        hook_B: MyHook = MyHook(self, 'Hook-B')
        hook_C: MyHook = MyHook(self, 'Hook-C')
        #
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_A.callback)
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_B.err_callback)
        hook_manager.add_hook(MyEvents.ACTION, MyHookable, hook_C.callback)
        #
        key = (MyEvents.CREATE, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        self.assertIn(hook_B.err_callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.ACTION, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_C.callback, hook_manager.hooks[key].callbacks)
        # Create event sources, emits CREATE
        self.output.clear()
        src_A: MySuperHookable = MySuperHookable(self, 'SuperSource-A')
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by SuperSource-A',
                              'SuperSource-A.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by SuperSource-A',
                              'SuperSource-A.CREATE hook call outcome: ERROR (Error in hook)'])
        self.output.clear()
        src_B: MySuperHookable = MySuperHookable(self, 'SuperSource-B')
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by SuperSource-B',
                              'SuperSource-B.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by SuperSource-B',
                              'SuperSource-B.CREATE hook call outcome: ERROR (Error in hook)'])
        # Install instance hooks
        hook_manager.add_hook(MyEvents.ACTION, src_A, hook_A.callback)
        hook_manager.add_hook(MyEvents.ACTION, src_B, hook_B.callback)
        #
        key = (MyEvents.ACTION, ANY, src_A)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.ACTION, ANY, src_B)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_B.callback, hook_manager.hooks[key].callbacks)
        # And action!
        self.output.clear()
        src_A.action()
        self.assertListEqual(self.output,
                             ['SuperSource-A.ACTION!',
                              'Hook Hook-A event ACTION called by SuperSource-A',
                              'SuperSource-A.ACTION hook call outcome: OK',
                              'Hook Hook-C event ACTION called by SuperSource-A',
                              'SuperSource-A.ACTION hook call outcome: OK'])
        #
        self.output.clear()
        src_B.action()
        self.assertListEqual(self.output,
                             ['SuperSource-B.ACTION!',
                              'Hook Hook-B event ACTION called by SuperSource-B',
                              'SuperSource-B.ACTION hook call outcome: OK',
                              'Hook Hook-C event ACTION called by SuperSource-B',
                              'SuperSource-B.ACTION hook call outcome: OK'])
    def test_inheritance(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        hook_manager.register_class(MySuperHookable, ('super-action', ))
        self.assertTupleEqual(tuple(hook_manager.hookables.keys()), (MyHookable, MySuperHookable))
        self.assertSetEqual(hook_manager.hookables[MyHookable],
                            set(x for x in cast(Enum, MyEvents).__members__.values()))
        self.assertTupleEqual(hook_manager.hookables[MySuperHookable], ('super-action', ))
        # Install hooks
        hook_A: MyHook = MyHook(self, 'Hook-A')
        hook_B: MyHook = MyHook(self, 'Hook-B')
        hook_C: MyHook = MyHook(self, 'Hook-C')
        hook_S: MyHook = MyHook(self, 'Hook-S')
        #
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_A.callback)
        hook_manager.add_hook(MyEvents.CREATE, MySuperHookable, hook_B.err_callback)
        hook_manager.add_hook(MyEvents.ACTION, MyHookable, hook_C.callback)
        hook_manager.add_hook('super-action', MySuperHookable, hook_S.callback)
        #
        key = (MyEvents.CREATE, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.CREATE, MySuperHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_B.err_callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.ACTION, MyHookable, ANY)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_C.callback, hook_manager.hooks[key].callbacks)
        # Create event sources, emits CREATE
        self.output.clear()
        src_A: MySuperHookable = MySuperHookable(self, 'SuperSource-A')
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by SuperSource-A',
                              'SuperSource-A.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by SuperSource-A',
                              'SuperSource-A.CREATE hook call outcome: ERROR (Error in hook)'])
        self.output.clear()
        src_B: MySuperHookable = MySuperHookable(self, 'SuperSource-B')
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by SuperSource-B',
                              'SuperSource-B.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by SuperSource-B',
                              'SuperSource-B.CREATE hook call outcome: ERROR (Error in hook)'])
        # Install instance hooks
        hook_manager.add_hook(MyEvents.ACTION, src_A, hook_A.callback)
        hook_manager.add_hook(MyEvents.ACTION, src_B, hook_B.callback)
        #
        key = (MyEvents.ACTION, ANY, src_A)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_A.callback, hook_manager.hooks[key].callbacks)
        key = (MyEvents.ACTION, ANY, src_B)
        self.assertTrue(key in hook_manager.hooks)
        self.assertIn(hook_B.callback, hook_manager.hooks[key].callbacks)
        # And action!
        self.output.clear()
        src_A.action()
        self.assertListEqual(self.output,
                             ['SuperSource-A.ACTION!',
                              'Hook Hook-A event ACTION called by SuperSource-A',
                              'SuperSource-A.ACTION hook call outcome: OK',
                              'Hook Hook-C event ACTION called by SuperSource-A',
                              'SuperSource-A.ACTION hook call outcome: OK'])
        #
        self.output.clear()
        src_B.action()
        self.assertListEqual(self.output,
                             ['SuperSource-B.ACTION!',
                              'Hook Hook-B event ACTION called by SuperSource-B',
                              'SuperSource-B.ACTION hook call outcome: OK',
                              'Hook Hook-C event ACTION called by SuperSource-B',
                              'SuperSource-B.ACTION hook call outcome: OK'])
        #
        self.output.clear()
        src_B.super_action()
        self.assertListEqual(self.output,
                             ['SuperSource-B.SUPER-ACTION!',
                              'Hook Hook-S event super-action called by SuperSource-B',
                              'SuperSource-B.SUPER-ACTION hook call outcome: OK'])
    def test_bad_hooks(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        hook_manager.register_class(MySuperHookable, ('super-action', ))
        src_A: MyHookable = MyHookable(self, 'Source-A')
        src_B: MySuperHookable = MySuperHookable(self, 'SuperSource-B')
        # Install hooks
        bad_hook: MyHook = MyHook(self, 'BAD-Hook')
        # Wrong hookables
        with self.assertRaises(TypeError) as cm:
            hook_manager.add_hook(MyEvents.CREATE, ANY, bad_hook.callback) # hook object
        self.assertEqual(cm.exception.args, ("Subject must be hookable class or instance, or name",))
        with self.assertRaises(TypeError) as cm:
            hook_manager.add_hook(MyEvents.CREATE, Enum, bad_hook.callback) # hook class
        self.assertEqual(cm.exception.args, ('The type is not registered as hookable',))
        self.assertDictEqual(hook_manager.hooks._reg, {})
        # Wrong events
        with self.assertRaises(ValueError) as cm:
            hook_manager.add_hook('BAD EVENT', MyHookable, bad_hook.callback)
        self.assertEqual(cm.exception.args, ("Event 'BAD EVENT' is not supported by 'MyHookable'",))
        with self.assertRaises(ValueError) as cm:
            hook_manager.add_hook('BAD EVENT', MySuperHookable, bad_hook.callback)
        self.assertEqual(cm.exception.args, ("Event 'BAD EVENT' is not supported by 'MySuperHookable'",))
        #
        with self.assertRaises(ValueError) as cm:
            hook_manager.add_hook('BAD EVENT', src_A, bad_hook.callback)
        self.assertEqual(cm.exception.args, ("Event 'BAD EVENT' is not supported by 'MyHookable'",))
        with self.assertRaises(ValueError) as cm:
            hook_manager.add_hook('BAD EVENT', src_B, bad_hook.callback)
        self.assertEqual(cm.exception.args, ("Event 'BAD EVENT' is not supported by 'MySuperHookable'",))
        # Bad hookable instances
        with self.assertRaises(TypeError) as cm:
            hook_manager.register_name(self, 'BAD_CLASS')
        self.assertEqual(cm.exception.args, ("The instance is not of hookable type",))
    def test_any_event(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        # Install hooks
        hook_A: MyHook = MyHook(self, 'Hook-A')
        hook_B: MyHook = MyHook(self, 'Hook-B')
        hook_C: MyHook = MyHook(self, 'Hook-C')
        hook_D: MyHook = MyHook(self, 'Hook-D')
        hook_manager.add_hook(ANY, MyHookable, hook_A.callback)
        hook_manager.add_hook(ANY, MyHookable, hook_B.err_callback)
        # Create event sources, emits CREATE
        self.output.clear()
        src_A: MyHookable = MyHookable(self, 'Source-A', register=True)
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: ERROR (Error in hook)'])
        # Install instance hooks
        hook_manager.add_hook(ANY, src_A, hook_C.callback)
        hook_manager.add_hook(ANY, 'Source-A', hook_D.callback)
        # And action!
        self.output.clear()
        src_A.action()
        self.assertListEqual(self.output,
                             ['Source-A.ACTION!',
                              'Hook Hook-C event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: OK',
                              'Hook Hook-D event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: OK',
                              'Hook Hook-A event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: OK',
                              'Hook Hook-B event ACTION called by Source-A',
                              'Source-A.ACTION hook call outcome: ERROR (Error in hook)'])
        # Optimizations
        self.assertIn(HookFlag.CLASS, hook_manager.flags)
        self.assertIn(HookFlag.INSTANCE, hook_manager.flags)
        self.assertIn(HookFlag.ANY_EVENT, hook_manager.flags)
        self.assertIn(HookFlag.NAME, hook_manager.flags)
    def test_class_hooks(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        # Install hooks
        hook_A: MyHook = MyHook(self, 'Hook-A')
        hook_B: MyHook = MyHook(self, 'Hook-B')
        hook_C: MyHook = MyHook(self, 'Hook-C')
        hook_D: MyHook = MyHook(self, 'Hook-D')
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_A.callback)
        hook_manager.add_hook(ANY, MyHookable, hook_B.err_callback)
        hook_manager.add_hook(MyEvents.CREATE, 'Source-A', hook_C.callback)
        hook_manager.add_hook(ANY, 'Source-A', hook_D.callback)
        # Create event sources, emits CREATE
        self.output.clear()
        MyHookable(self, 'Source-A', use_class=True)
        self.assertListEqual(self.output,
                             ['Hook Hook-A event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: OK',
                              'Hook Hook-B event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: ERROR (Error in hook)'])
    def test_name_hooks(self):
        # register hookables
        hook_manager.register_class(MyHookable, MyEvents)
        # Install hooks
        hook_A: MyHook = MyHook(self, 'Hook-A')
        hook_B: MyHook = MyHook(self, 'Hook-B')
        hook_C: MyHook = MyHook(self, 'Hook-C')
        hook_D: MyHook = MyHook(self, 'Hook-D')
        hook_manager.add_hook(MyEvents.CREATE, MyHookable, hook_A.callback)
        hook_manager.add_hook(ANY, MyHookable, hook_B.err_callback)
        hook_manager.add_hook(MyEvents.CREATE, 'Source-A', hook_C.callback)
        hook_manager.add_hook(ANY, 'Source-A', hook_D.err_callback)
        # Create event sources, emits CREATE
        self.output.clear()
        MyHookable(self, 'Source-A', use_name=True)
        self.assertListEqual(self.output,
                             ['Hook Hook-C event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: OK',
                              'Hook Hook-D event CREATE called by Source-A',
                              'Source-A.CREATE hook call outcome: ERROR (Error in hook)'])



if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: firebird/base/base_test.proto

from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='firebird/base/base_test.proto',
  package='firebird.base',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x1d\x66irebird/base/base_test.proto\x12\rfirebird.base\x1a\x19google/protobuf/any.proto\x1a\x1cgoogle/protobuf/struct.proto\"@\n\tTestState\x12\x0c\n\x04name\x18\x01 \x01(\t\x12%\n\x04test\x18\x02 \x01(\x0e\x32\x17.firebird.base.TestEnum\"\xc8\x01\n\x0eTestCollection\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\'\n\x05tests\x18\x02 \x03(\x0b\x32\x18.firebird.base.TestState\x12(\n\x07\x63ontext\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12+\n\nannotation\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12(\n\nsupplement\x18\x05 \x03(\x0b\x32\x14.google.protobuf.Any*\xd8\x01\n\x08TestEnum\x12\x10\n\x0cTEST_UNKNOWN\x10\x00\x12\x0e\n\nTEST_READY\x10\x01\x12\x10\n\x0cTEST_RUNNING\x10\x02\x12\x10\n\x0cTEST_WAITING\x10\x03\x12\x12\n\x0eTEST_SUSPENDED\x10\x04\x12\x11\n\rTEST_FINISHED\x10\x05\x12\x10\n\x0cTEST_ABORTED\x10\x06\x12\x10\n\x0cTEST_CREATED\x10\x01\x12\x10\n\x0cTEST_BLOCKED\x10\x03\x12\x10\n\x0cTEST_STOPPED\x10\x04\x12\x13\n\x0fTEST_TERMINATED\x10\x06\x1a\x02\x10\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])

_TESTENUM = _descriptor.EnumDescriptor(
  name='TestEnum',
  full_name='firebird.base.TestEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TEST_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_READY', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_RUNNING', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_WAITING', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_SUSPENDED', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_FINISHED', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_ABORTED', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_CREATED', index=7, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_BLOCKED', index=8, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_STOPPED', index=9, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEST_TERMINATED', index=10, number=6,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=b'\020\001',
  serialized_start=375,
  serialized_end=591,
)
_sym_db.RegisterEnumDescriptor(_TESTENUM)

TestEnum = enum_type_wrapper.EnumTypeWrapper(_TESTENUM)
TEST_UNKNOWN = 0
TEST_READY = 1
TEST_RUNNING = 2
TEST_WAITING = 3
TEST_SUSPENDED = 4
TEST_FINISHED = 5
TEST_ABORTED = 6
TEST_CREATED = 1
TEST_BLOCKED = 3
TEST_STOPPED = 4
TEST_TERMINATED = 6



_TESTSTATE = _descriptor.Descriptor(
  name='TestState',
  full_name='firebird.base.TestState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='firebird.base.TestState.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test', full_name='firebird.base.TestState.test', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=169,
)


_TESTCOLLECTION = _descriptor.Descriptor(
  name='TestCollection',
  full_name='firebird.base.TestCollection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='firebird.base.TestCollection.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tests', full_name='firebird.base.TestCollection.tests', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='context', full_name='firebird.base.TestCollection.context', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='annotation', full_name='firebird.base.TestCollection.annotation', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supplement', full_name='firebird.base.TestCollection.supplement', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=172,
  serialized_end=372,
)

_TESTSTATE.fields_by_name['test'].enum_type = _TESTENUM
_TESTCOLLECTION.fields_by_name['tests'].message_type = _TESTSTATE
_TESTCOLLECTION.fields_by_name['context'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_TESTCOLLECTION.fields_by_name['annotation'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_TESTCOLLECTION.fields_by_name['supplement'].message_type = google_dot_protobuf_dot_any__pb2._ANY
DESCRIPTOR.message_types_by_name['TestState'] = _TESTSTATE
DESCRIPTOR.message_types_by_name['TestCollection'] = _TESTCOLLECTION
DESCRIPTOR.enum_types_by_name['TestEnum'] = _TESTENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TestState = _reflection.GeneratedProtocolMessageType('TestState', (_message.Message,), {
  'DESCRIPTOR' : _TESTSTATE,
  '__module__' : 'firebird.base.base_test_pb2'
  # @@protoc_insertion_point(class_scope:firebird.base.TestState)
  })
_sym_db.RegisterMessage(TestState)

TestCollection = _reflection.GeneratedProtocolMessageType('TestCollection', (_message.Message,), {
  'DESCRIPTOR' : _TESTCOLLECTION,
  '__module__' : 'firebird.base.base_test_pb2'
  # @@protoc_insertion_point(class_scope:firebird.base.TestCollection)
  })
_sym_db.RegisterMessage(TestCollection)


_TESTENUM._options = None
# @@protoc_insertion_point(module_scope)



#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/logging.py
# DESCRIPTION:    Context-based logging
# CREATED:        14.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""firebird-base - Context-based logging

"""

from __future__ import annotations
from typing import Any, Dict, Tuple, Union, Hashable
from enum import IntEnum, Flag, auto
from collections.abc import Mapping
from dataclasses import dataclass
from logging import Logger, LoggerAdapter, getLogger, lastResort, Formatter
from .types import UNDEFINED, DEFAULT, ANY, ALL, Distinct, CachedDistinct, Sentinel
from .collections import Registry

class LogLevel(IntEnum):
    "Shadow enumeration for logging levels"
    NOTSET = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    FATAL = CRITICAL
    WARN = WARNING

class BindFlag(Flag):
    "Internal flags used by `LoggingManager`."
    DIRECT = auto()
    ANY_AGENT = auto()
    ANY_CTX = auto()
    ANY_ANY = auto()

class FBLoggerAdapter(LoggerAdapter, CachedDistinct):
    """`~logging.LoggerAdapter` that injects information about context, agent and topic
into `extra` and with **f-string** log message support.

Attributes:
    logger (Logger): Adapted Logger instance.
    agent (str): Agent for logger
    context (str): Context for logger
    topic (str): Topic for logger.
"""
    def __init__(self, logger: Logger, agent: Any=UNDEFINED, context: Any=UNDEFINED, topic: str=''):
        """
Arguments:
    logger: Adapted Logger instance.
    agent: Agent for logger
    context: Context for logger
    topic: Topic of recorded information.
"""
        self.logger: Logger = logger
        self.agent: Any = agent
        self.context: Any = context
        self.topic: str = topic
    @classmethod
    def extract_key(cls, *args, **kwargs) -> Hashable:
        "Returns instance key extracted from constructor arguments."
        return (args[1], args[2])
    def get_key(self) -> Hashable: # pragma: no cover
        "Returns instance key."
        return (self.topic, self.agent, self.context)
    def process(self, msg, kwargs) -> Tuple[str, Dict]:
        """Process the logging message and keyword arguments passed into
        a logging call to insert contextual information. You can either
        manipulate the message itself, the keyword args or both. Return
        the message and kwargs modified (or not) to suit your needs.
        """
        return msg, kwargs
    def log(self, level, msg, *args, **kwargs):
        """Delegate a log call to the underlying logger after processing.

        Interpolates the message as **f-string** using either `kwargs` or dict passed as
        only one positional argument. If sole positional argument is not dictionary or
        `args` has more than one item, adds `args` into namespace for interpolation.

        Moves 'context', 'agent' and 'topic' keyword arguments into `extra`.

        Strips out all keyword arguments not expected by `logging.Logger`.
        """
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            if (args and len(args) == 1 and isinstance(args[0], Mapping) and args[0]):
                ns = args[0]
            else:
                ns = kwargs
                if args:
                    ns['args'] = args
            msg = eval(f'f"""{msg}"""', globals(), ns)
            args = ()
            if 'stacklevel' not in kwargs:
                kwargs['stacklevel'] = 3
            kwargs.setdefault('extra', {}).update(topic=self.topic, agent=self.agent,
                                                  context=self.context)
            self.logger.log(level, msg, *args, **{k: v for k, v in kwargs.items()
                                                  if k in ['exc_info', 'stack_info',
                                                           'stacklevel', 'extra']})

@dataclass(order=True, frozen=True)
class BindInfo(Distinct):
    """Information about Logger binding"""
    topic: str
    agent: str
    context: str
    logger: FBLoggerAdapter
    def get_key(self) -> Any:
        return (self.topic, self.agent, self.context)

def get_logging_id(obj: Any) -> Any:
    """Returns logging ID for object.

Arguments:
    obj: Any object

Returns:
    1. `logging_id` attribute if `obj` does have it, or..
    2. `__name__` attribute if `obj` does have it, or..
    3. `str(obj)`
"""
    return getattr(obj, 'logging_id', getattr(obj, '__name__', str(obj)))

class LoggingIdMixin:
    "Mixin class that adds `logging_id` property and `__str__`."
    def __str__(self):
        "Returns `logging_id`"
        return self.logging_id
    @property
    def logging_id(self) -> str:
        "Returns `_logging_id_` attribute if defined, else returns qualified class name"
        return getattr(self, '_logging_id_', self.__class__.__qualname__)

class LoggingManager:
    """Logger manager.
"""
    def __init__(self):
        self.loggers: Registry = Registry()
        self.topics: Dict = {}
        self.bindings: BindFlag = BindFlag(0)
    def _update_bindings(self, agent: Any, context: Any) -> None:
        if agent is ANY:
            self.bindings |= BindFlag.ANY_AGENT
        if context is ANY:
            self.bindings |= BindFlag.ANY_CTX
        if (agent is ANY) and (context is ANY):
            self.bindings |= BindFlag.ANY_ANY
        if (agent is not ANY) and (context is not ANY):
            self.bindings |= BindFlag.DIRECT
    def _update_topics(self, topic: str) -> None:
        if topic in self.topics:
            self.topics[topic] += 1
        else:
            self.topics[topic] = 1
    def bind_logger(self, agent: Any, context: Any, logger: Union[str, Logger], topic: str='') -> None:
        """Bind agent and context to specific logger.

Arguments:
    agent: Agent identification
    context: Context identification
    logger: Loger (instance or name)
    topic: Topic of recorded information

The identification of agent and context could be:

    1. String
    2. Object instance. Uses `get_logging_id()` to retrieve its logging ID.
    3. Sentinel. The ANY sentinel matches any particular agent or context. You can use
       sentinel `.UNDEFINED` to register a logger for cases when agent or context are not
       specified in logger lookup.

Important:
   You SHOULD NOT use sentinel `.ALL` for `agent` or `context` identification! This
   sentinel is used by `.unbind`, so bindings that use ALL could not be removed by `.unbind`.

"""
        if isinstance(logger, str):
            logger = getLogger(logger)
        if not isinstance(agent, (str, Sentinel)):
            agent = get_logging_id(agent)
        if not isinstance(context, (str, Sentinel)):
            context = get_logging_id(context)
        if agent is not ANY and context is not ANY:
            logger = FBLoggerAdapter(logger, agent, context)
        self._update_bindings(agent, context)
        self._update_topics(topic)
        self.loggers.update(BindInfo(topic, agent, context, logger))
    def unbind(self, agent: Any, context: Any, topic: str='') -> int:
        """Drops logger bindings.
"""
        if not isinstance(agent, (str, Sentinel)):
            agent = get_logging_id(agent)
        if not isinstance(context, (str, Sentinel)):
            context = get_logging_id(context)
        if topic in self.topics:
            rm = [i for i in self.loggers
                  if i.topic == topic and ((i.agent == agent) or agent is ALL)
                  and ((i.context == context) or context is ALL)]
            for item in rm:
                self.loggers.remove(item)
            # recalculate optimizations
            self.topics.clear()
            self.bindings = BindFlag(0)
            for item in self.loggers:
                self._update_bindings(item.agent, item.context)
                self._update_topics(item.topic)
            return len(rm)
        return 0
    def clear(self) -> None:
        """Remove all logger bindings"""
        self.loggers.clear()
        self.topics.clear()
        self.bindings = BindFlag(0)
    def get_logger(self, agent: Any=UNDEFINED, context: Any=DEFAULT, topic: str='') -> FBLoggerAdapter:
        """Return a logger for the specified agent and context combination.

Arguments:
    agent: Agent identification.
    context: Context identification.
    topic: Topic of recorded information.

The identification of agent and context could be:

    1. String
    2. Object instance. Uses `get_logging_id()` to retrieve its logging ID.
    3. Sentinel `.UNDEFINED`
    4. When `context` is sentinel `.DEFAULT`, uses `agent` attribute `log_context` (if defined)
       or sentinel `.UNDEFINED` otherwise.

The search for a suitable topic logger proceeds as follows:

    1. Return logger registered for specified agent and context, or...
    2. Return logger registered for ANY agent and specified context, or...
    3. Return logger registered for specified agent and ANY context, or...
    4. Return logger registered for ANY agent and ANY context, or...
    5. Return the root logger.
"""
        if context is DEFAULT:
            context = getattr(agent, 'log_context', UNDEFINED)
        if agent is not UNDEFINED and not isinstance(agent, str):
            agent = get_logging_id(agent)
        if context is not UNDEFINED and not isinstance(context, str):
            context = get_logging_id(context)
        result: BindInfo = None
        if topic in self.topics:
            if BindFlag.DIRECT in self.bindings and \
               (result := self.loggers.get((topic, agent, context))) is not None:
                result = result.logger
            elif BindFlag.ANY_AGENT in self.bindings and \
                 (result := self.loggers.get((topic, ANY, context))) is not None:
                result = result.logger
            elif BindFlag.ANY_CTX in self.bindings and \
                 (result := self.loggers.get((topic, agent, ANY))) is not None:
                result = result.logger
            elif BindFlag.ANY_ANY in self.bindings and \
                 (result := self.loggers.get((topic, ANY, ANY))) is not None:
                result = result.logger
            else:
                result = getLogger(topic)
        else:
            result = getLogger(topic)
        return result if isinstance(result, FBLoggerAdapter) \
               else FBLoggerAdapter(result, agent, context, topic)

#: Logging Manager
logging_manager: LoggingManager = LoggingManager()

bind_logger = logging_manager.bind_logger
get_logger = logging_manager.get_logger

# Install simple formatter for lastResort handler
if lastResort is not None and lastResort.formatter is None:
    lastResort.setFormatter(Formatter('%(levelname)s: %(message)s'))

def install_null_logger():
    "Installs 'null' logger."
    log = getLogger('null')
    log.propagate = False
    log.disabled = True




#coding:utf-8
#
# PROGRAM/MODULE: firebird-base
# FILE:           firebird/base/config.py
# DESCRIPTION:    Classes for configuration definitions
# CREATED:        14.5.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Firebird Base - Classes for configuration definitions

Complex applications (and some library modules like `logging`) could be often parametrized
via configuration. This module provides a framework for unified structured configuration
that supports:

* configuration options of various data type, including lists and other complex types
* validation
* direct manipulation of configuration values
* reading from (and writing into) configuration in `configparser` format
* exchanging configuration (for example between processes) using Google protobuf messages
"""

from __future__ import annotations
from typing import Generic, Type, Any, List, Dict, Union, Sequence, Optional, TypeVar, cast, get_type_hints
from abc import ABC, abstractmethod
from uuid import UUID
from decimal import Decimal, DecimalException
from configparser import ConfigParser
from inspect import signature, Signature, Parameter
from enum import Enum, Flag, _decompose
from .config_pb2 import ConfigProto
from .types import Error, MIME, ZMQAddress, PyExpr, PyCode, PyCallable
from .strconv import get_convertor, convert_to_str, Convertor

# Functions

def unindent_verticals(value: str) -> str:
    lines = []
    indent = None
    for line in value.split('\n'):
        if line.startswith('|'):
            if indent is None:
                indent = (len(line[1:]) - len(line[1:].strip())) + 1
            lines.append(line[indent:])
        else:
            lines.append(line)
    return '\n'.join(lines)

def _eq(a: Any, b: Any) -> bool:
    return str(a) == str(b)

def create_config(_cls: Type[Config], name: str, description: str) -> Config: # pragma: no cover
    """Return newly created `Config` instance. Intended to be used with `functools.partial`.
"""
    return _cls(name, description)



# Base classes

T = TypeVar('T')

class Option(Generic[T], ABC):
    """Generic abstract base class for configuration options.

Arguments:
    name: Option name.
    datatype: Option datatype.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.

Attributes:
    name (str): Option name.
    datatype (type): Option datatype.
    description (str): Option description. Can span multiple lines.
    required (bool): True if option must have a value.
    default (instance of [T]): Default option value.
"""
    def __init__(self, name: str, datatype: T, description: str, required: bool=False,
                 default: T=None):
        assert name and isinstance(name, str), "name required"
        assert datatype and isinstance(datatype, type), "datatype required"
        assert description and isinstance(description, str), "description required"
        assert default is None or isinstance(default, datatype), "default has wrong data type"
        self.name: str = name
        self.datatype: T = datatype
        self.description: str = description
        self.required: bool = required
        self.default: T = default
        if default is not None:
            self.set_value(default)
    def _check_value(self, value: T) -> None:
        if value is None and self.required:
            raise ValueError(f"Value is required for option '{self.name}'.")
        if value is not None and not isinstance(value, self.datatype):
            raise TypeError(f"Option '{self.name}' value must be a "
                            f"'{self.datatype.__name__}',"
                            f" not '{type(value).__name__}'")
    def _get_config_lines(self) -> List[str]:
        """Returns list of strings containing text lines suitable for use in configuration
file processed with `~configparser.ConfigParser`.

Text lines with configuration start with comment marker `;` and end with newline.

Note:
   This function is intended for internal use. To get string describing current
   configuration that is suitable for configuration files, use `get_config` method.
"""
        lines = [f"; {self.name}\n",
                 f"; {'-' * len(self.name)}\n",
                 ";\n",
                 f"; data type: {self.datatype.__name__}\n",
                 ";\n"]
        if self.required:
            description = '[REQUIRED] ' + self.description
        else:
            description = '[optional] ' + self.description
        for line in description.split('\n'):
            lines.append(f"; {line}\n")
        lines.append(';\n')
        value = self.get_value()
        nodef = ';' if value == self.default else ''
        value = '<UNDEFINED>' if value is None else self.get_formatted()
        if '\n' in value:
            chunks = value.splitlines(keepends=True)
            new_value = [chunks[0]]
            new_value.extend(f'{nodef}{x}' for x in chunks[1:])
            value = ''.join(new_value)
        lines.append(f"{nodef}{self.name} = {value}\n")
        return lines
    def load_config(self, config: ConfigParser, section: str) -> None:
        """Update option value from `~configparser.ConfigParser` instance.

Arguments:
    config:  ConfigParser instance.
    section: Name of ConfigParser section that should be used to get new option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        if self.name in config[section]:
            self.set_as_str(config[section][self.name])
    def validate(self) -> None:
        """Validates option state.

Raises:
    Error: When required option does not have a value.
"""
        if self.required and self.get_value() is None:
            raise Error(f"Missing value for required option '{self.name}'")
    def get_config(self) -> str:
        """Returns string containing text lines suitable for use in configuration file
processed with `~configparser.ConfigParser`.

Text lines with configuration start with comment marker ; and end with newline.
"""
        return ''.join(self._get_config_lines())
    @abstractmethod
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
    @abstractmethod
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
    @abstractmethod
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
    @abstractmethod
    def get_as_str(self) -> str:
        """Return value as string."""
    @abstractmethod
    def get_value(self) -> T:
        "Return current option value"
    @abstractmethod
    def set_value(self, value: T) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
    @abstractmethod
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
    @abstractmethod
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""

class Config:
    """Collection of configuration options.

Arguments:
    name: Name associated with Config (default section name).

Important:
    Descendants must define individual options and sub configs as instance attributes.
"""
    def __init__(self, name: str):
        self._name: str = name
    def validate(self) -> None:
        """Checks whether:
    - all required options have value other than None.
    - all options are defined as config attribute with the same name as option name

Raises exception when any constraint required by configuration is violated.
"""
        for option in self.options:
            option.validate()
            if not hasattr(self, option.name):
                raise Error(f"Option '{option.name}' is not defined as "
                            f"attribute with the same name")
    def clear(self, to_default: bool=True) -> None:
        """Clears all owned options and options in owned sub-configs.

Arguments:
    to_default: If True, sets the option values to defaults, else to None.
"""
        for option in self.options:
            option.clear(to_default)
        for config in self.configs:
            config.clear(to_default)
    def get_description(self) -> str:
        """Configuration description. Can span multiple lines.

Note:  Default implementation returns class doc string.
"""
        return self.__doc__
    def get_config(self) -> str:
        """Returns string containing text lines suitable for use in configuration file
processed with `~configparser.ConfigParser`.

Text lines with configuration start with comment marker ; and end with newline.
"""
        lines = [f'[{self.name}]\n', ';\n']
        for line in self.get_description().splitlines():
            lines.append(f"; {line}\n")
        lines.append(';\n')
        for option in self.options:
            lines.append('\n')
            lines.append(option.get_config())
        for config in self.configs:
            lines.append('\n')
            lines.append(config.get_config())
        return ''.join(lines)
    def load_config(self, config: ConfigParser, section: str=None) -> None:
        """Update configuration.

Arguments:
    config:  ConfigParser instance with configuration values.
    section: Name of ConfigParser section that should be used to get new configuration
             values. If not provided, uses `name`.
"""
        if section is None:
            section = self.name
        try:
            for option in self.options:
                option.load_config(config, section)
            for subcfg in self.configs:
                subcfg.load_config(config)
        except Exception as exc: # pragma: no cover
            raise Error(f"Configuration error: {exc}") from exc
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains option values and sub-configs.
"""
        for option in self.options:
            option.load_proto(proto)
        for subcfg in self.configs:
            if subcfg.name in proto.configs:
                subcfg.load_proto(proto.configs[subcfg.name])
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option values and sub-configs should be stored.
"""
        for option in self.options:
            option.save_proto(proto)
        for subcfg in self.configs:
            subcfg.save_proto(proto.configs.get_or_create(subcfg.name))
    @property
    def name(self) -> str:
        "Name associated with Config (default section name)."
        return self._name
    @property
    def options(self) -> List[Option]:
        "List of options defined for this Config instance."
        return [v for v in vars(self).values() if isinstance(v, Option)]
    @property
    def configs(self) -> List[Config]:
        """List of sub-Configs defined for this Config instance. It includes all instance
attributes of `Config` type, and `Config` values of owned `ConfigOption` and `ConfigListOption`
instances.
"""
        result = [v if isinstance(v, Config) else v.value
                  for v in vars(self).values() if isinstance(v, (Config, ConfigOption))]
        for opt in (v for v in vars(self).values() if isinstance(v, ConfigListOption)):
            result.extend(opt.value)
        return result

# Options

class StrOption(Option[str]):
    """Configuration option with string value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: str=None):
        self._value: str = None
        super().__init__(name, str, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = self._value
        if '\n' in result:
            lines = []
            for line in result.splitlines(True):
                if lines:
                    lines.append('   ' + line)
                else:
                    lines.append(line)
            result = ''.join(lines)
        return result
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = value
    def get_as_str(self) -> str:
        """Return value as string."""
        return self._value
    def get_value(self) -> str:
        "Return current option value"
        return self._value
    def set_value(self, value: str) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_value(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value
    value: str = property(get_value, set_value)

class IntOption(Option[int]):
    """Configuration option with integer value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
    signed: When False, the option value cannot be negative.
"""
    def __init__(self, name: str, description: str, *, required: bool=False,
                 default: int=None, signed: bool=False):
        self._value: int = None
        self.__signed: bool = signed
        super().__init__(name, int, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else str(self._value)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        new = int(value)
        if not self.__signed and new < 0:
            raise ValueError("Negative numbers not allowed")
        self._value = new
    def get_as_str(self) -> str:
        """Return value as string."""
        return str(self._value)
    def get_value(self) -> int:
        "Return current option value"
        return self._value
    def set_value(self, value: int) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        if value is not None and (not self.__signed and value < 0):
            raise ValueError("Negative numbers not allowed")
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""

        if self.name in proto.options:
            opt = proto.options[self.name]
            oneof = opt.WhichOneof('kind')
            if oneof not in ['as_sint32', 'as_sint64', 'as_uint32', 'as_uint64', 'as_string']:
                raise TypeError(f"Wrong value type: {oneof[3:]}")
            if oneof == 'as_string':
                self.set_as_str(opt.as_string)
            else:
                self.set_value(getattr(opt, oneof))
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            opt = proto.options[self.name]
            if self.__signed:
                opt.as_sint64 = self._value
            else:
                opt.as_uint64 = self._value
    value: int = property(get_value, set_value)

class FloatOption(Option[float]):
    """Configuration option with float value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: float=None):
        self._value: float = None
        super().__init__(name, float, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else str(self._value)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = float(value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return str(self._value)
    def get_value(self) -> float:
        "Return current option value"
        return self._value
    def set_value(self, value: float) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""

        if self.name in proto.options:
            opt = proto.options[self.name]
            oneof = opt.WhichOneof('kind')
            if oneof not in ['as_float', 'as_double', 'as_string']:
                raise TypeError(f"Wrong value type: {oneof[3:]}")
            if oneof == 'as_string':
                self.set_as_str(opt.as_string)
            else:
                self.set_value(getattr(opt, oneof))
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_double = self._value
    value: float = property(get_value, set_value)

class DecimalOption(Option[Decimal]):
    """Configuration option with decimal.Decimal value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: Decimal=None):
        self._value: Decimal = None
        super().__init__(name, Decimal, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else str(self._value)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        try:
            self._value = Decimal(value)
        except DecimalException as exc:
            raise ValueError(str(exc))
    def get_as_str(self) -> str:
        """Return value as string."""
        return str(self._value)
    def get_value(self) -> Decimal:
        "Return current option value"
        return self._value
    def set_value(self, value: Decimal) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto):
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""

        if self.name in proto.options:
            opt = proto.options[self.name]
            oneof = opt.WhichOneof('kind')
            if oneof not in ['as_sint32', 'as_sint64', 'as_uint32', 'as_uint64', 'as_string']:
                raise TypeError(f"Wrong value type: {oneof[3:]}")
            if oneof == 'as_string':
                self.set_as_str(opt.as_string)
            else:
                self.set_value(Decimal(getattr(opt, oneof)))
    def save_proto(self, proto: ConfigProto):
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = str(self._value)
    value: Decimal = property(get_value, set_value)

class BoolOption(Option[bool]):
    """Configuration option with boolean value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: bool=None):
        self._value: bool = None
        self.from_str = get_convertor(bool).from_str
        super().__init__(name, bool, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        return 'yes' if self._value else 'no'
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = self.from_str(bool, value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return str(self._value)
    def get_value(self) -> bool:
        "Return current option value"
        return self._value
    def set_value(self, value: bool) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            oneof = opt.WhichOneof('kind')
            if oneof not in ['as_bool', 'as_string']:
                raise TypeError(f"Wrong value type: {oneof[3:]}")
            if oneof == 'as_string':
                self.set_as_str(opt.as_string)
            else:
                self.set_value(opt.as_bool)
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_bool = self._value
    value: bool = property(get_value, set_value)

class ZMQAddressOption(Option[ZMQAddress]):
    """Configuration option with `.ZMQAddress` value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False,
                 default: ZMQAddress=None):
        self._value: ZMQAddress = None
        super().__init__(name, ZMQAddress, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else self._value
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = ZMQAddress(value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return self._value
    def get_value(self) -> ZMQAddress:
        "Return current option value"
        return self._value
    def set_value(self, value: ZMQAddress) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value
    value: ZMQAddress = property(get_value, set_value)

class EnumOption(Option[Enum]):
    """Configuration option with enum value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
    allowed: List of allowed Enum members. When not defined, all members of enum type are
             allowed.

Attributes:
    allowed: List of allowed enum values.
"""
    def __init__(self, name: str, enum_class: Enum, description: str, *, required: bool=False,
                 default: Enum=None, allowed: List=None):
        self._value: Enum = None
        self.allowed: Sequence = enum_class if allowed is None else allowed
        self._members: Dict = {i.name.lower(): i for i in self.allowed}
        super().__init__(name, enum_class, description, required, default)
    def get_config(self) -> str:
        """Returns string containing text lines suitable for use in configuration file
processed with `~configparser.ConfigParser`.

Text lines with configuration start with comment marker ; and end with newline.
"""
        lines: List = super()._get_config_lines()
        lines.insert(4, f"; values: {', '.join(x.name.lower() for x in self.allowed)}\n")
        return ''.join(lines)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else self._value.name.lower()
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        name = value.lower()
        if name in self._members:
            self.set_value(self._members[name])
        else:
            raise ValueError(f"Illegal value '{value}' for enum type "
                             f"'{self.datatype.__name__}'")
    def get_as_str(self) -> str:
        """Return value as string."""
        return self._value.name
    def get_value(self) -> Enum:
        "Return current option value"
        return self._value
    def set_value(self, value: Enum) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        if value is not None and value not in self.allowed:
            raise ValueError(f"Value '{value}' not allowed")
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value.name
    value: Enum = property(get_value, set_value)

class FlagOption(Option[Flag]):
    """Configuration option with flag value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
    allowed: List of allowed Flag members. When not defined, all members of flag type are
             allowed.

Attributes:
    allowed: List of allowed flag values.
"""
    def __init__(self, name: str, flag_class: Flag, description: str, *, required: bool=False,
                 default: Flag=None, allowed: List=None):
        self._value: Flag = None
        self.allowed: Sequence = flag_class if allowed is None else allowed
        self._members: Dict = {i.name.lower(): i for i in self.allowed}
        super().__init__(name, flag_class, description, required, default)
    def get_config(self) -> str:
        """Returns string containing text lines suitable for use in configuration file
processed with `~configparser.ConfigParser`.

Text lines with configuration start with comment marker ; and end with newline.
"""
        lines: List = super()._get_config_lines()
        lines.insert(4, f"; values: {', '.join(x.name.lower() for x in self.allowed)}\n")
        return ''.join(lines)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else self.get_as_str().lower()
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        result = self.datatype(0)
        for name in (x.strip().lower() for x in value.split('|' if '|' in value else ',')):
            if name in self._members:
                result |= self._members[name]
            else:
                raise ValueError(f"Illegal value '{name}' for flag option '{self.name}'")
        self.set_value(result)
    def get_as_str(self) -> str:
        """Return value as string."""
        if self._value._name_ is not None:
            return self._value.name
        members, uncovered = _decompose(self.datatype, self._value)
        if len(members) == 1 and members[0]._name_ is None:
            return f'{members[0]._value_}'
        return ' | '.join([str(m._name_ or m._value_) for m in members])
    def get_value(self) -> Flag:
        "Return current option value"
        return self._value
    def set_value(self, value: Flag) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        if value is not None:
            members, uncovered = _decompose(self.datatype, value.value)
            if uncovered or [i for i in members if i.name is None or i.name.lower() not in self._members]:
                raise ValueError(f"Illegal value '{value!s}' for flag option '{self.name}'")
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            oneof = opt.WhichOneof('kind')
            if oneof not in ['as_uint64', 'as_string']:
                raise TypeError(f"Wrong value type: {oneof[3:]}")
            if oneof == 'as_uint64':
                self.set_value(self.datatype(opt.as_uint64))
            else:
                self.set_as_str(opt.as_string)
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_uint64 = self._value.value
    value: Flag = property(get_value, set_value)

class UUIDOption(Option[UUID]):
    """Configuration option with UUID value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: UUID=None):
        self._value: UUID = None
        super().__init__(name, UUID, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else str(self._value)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = UUID(value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return 'None' if self._value is None else self._value.hex
    def get_value(self) -> UUID:
        "Return current option value"
        return self._value
    def set_value(self, value: UUID) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            oneof = opt.WhichOneof('kind')
            if oneof not in ['as_bytes', 'as_string']:
                raise TypeError(f"Wrong value type: {oneof[3:]}")
            if oneof == 'as_bytes':
                self.set_value(UUID(bytes=opt.as_bytes))
            else:
                self.set_value(UUID(opt.as_string))
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_bytes = self._value.bytes
    value: UUID = property(get_value, set_value)

class MIMEOption(Option[MIME]):
    """Configuration option with MIME type specification value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: MIME=None):
        self._value: MIME = None
        super().__init__(name, MIME, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        return '<UNDEFINED>' if self._value is None else self._value
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = MIME(value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return 'None' if self._value is None else self._value
    def get_value(self) -> MIME:
        "Return current option value"
        return self._value
    def set_value(self, value: MIME) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value
    value: MIME = property(get_value, set_value)

class ListOption(Option[List]):
    """Configuration option with list of values.

Arguments:
    name:        Option name.
    item_type:   Datatype of list items. It could be a type or sequence of types. If multiple
                 types are provided, each value in config file must have format:
                 `type_name:value_as_str`.
    description: Option description. Can span multiple lines.
    required:    True if option must have a value.
    default:     Default option value.
    separator:   String that separates list item values when options value is read from
                 `ConfigParser`. It's possible to use a line break as separator.
                 If separator is `None` [default] and the value contains line breaks, it uses
                 the line break as separator, otherwise it uses comma as separator.

Attributes:
    item_types:  Datatypes of list items. If there is more than one type, each value in
                 config file must have format: `type_name:value_as_str`.
    separator:   String that separates list item values when options value is read from
                 `ConfigParser`. Default separator is None. It's possible to use a line
                 break as separator. If separator is `None` and the value contains line
                 breaks, it uses the line break as separator, otherwise it uses comma as
                 separator.

Important:
    When option is read from `ConfigParser`, empty values are ignored.
"""
    def __init__(self, name: str, item_type: Union[Type, Sequence[Type]], description: str,
                 *, required: bool=False, default: List=None, separator: str=None):
        self._value: List = None
        self.item_types: Sequence[Type] = (item_type, ) if isinstance(item_type, type) else item_type
        self.separator: Optional[str] = separator
        self._convertor: Convertor = get_convertor(item_type) if isinstance(item_type, type) else None
        super().__init__(name, list, description, required, default)
    def _get_as_typed_str(self, value: Any) -> str:
        result = convert_to_str(value)
        if len(self.item_types) > 1:
            result = f'{value.__class__.__name__}:{result}'
        return result
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = [convert_to_str(i) for i in self._value]
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        if sep == '\n':
            x = '\n   '
            return f"\n   {x.join(result)}"
        return f'{sep} '.join(result)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        new = []
        if value.strip():
            separator = ('\n' if '\n' in value else ',') if self.separator is None else self.separator
            itype = self.item_types[0]
            convertor = self._convertor
            name_map = {}
            if len(self.item_types) > 1:
                name_map = {cls.__name__: cls for cls in self.item_types}
                fullname_map = {f'{cls.__module__}.{cls.__name__}': cls for cls in self.item_types}
            for item in (i for i in value.split(separator) if i.strip()):
                if name_map:
                    itype_name, item = item.split(':', 1)
                    itype_name = itype_name.strip()
                    itype = fullname_map.get(itype_name) if '.' in itype_name else name_map.get(itype_name)
                    if itype is None:
                        raise ValueError(f"Item type '{itype_name}' not supported")
                    convertor = get_convertor(itype)
                new.append(convertor.from_str(itype, item.strip()))
            self._value = new
    def get_as_str(self) -> str:
        """Return value as string."""
        result = [convert_to_str(i) for i in self._value]
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        return sep.join(result)
    def get_value(self) -> List:
        "Return current option value"
        return self._value
    def set_value(self, value: List) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        if value is not None:
            i = 0
            for item in value:
                if item.__class__ not in self.item_types:
                    raise ValueError(f"List item[{i}] has wrong type")
                i += 1
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            result = [self._get_as_typed_str(i) for i in self._value]
            sep = self.separator
            if sep is None:
                sep = '\n' if sum(len(i) for i in result) > 80 else ','
            proto.options[self.name].as_string = sep.join(result)
    value: List = property(get_value, set_value)

class PyExprOption(Option[PyExpr]):
    """String configuration option with Python expression value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: PyExpr=None):
        self._value: PyExpr = None
        super().__init__(name, PyExpr, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = self._value
        if '\n' in result:
            lines = []
            for line in result.splitlines(True):
                if lines:
                    lines.append('   ' + line)
                else:
                    lines.append(line)
            result = ''.join(lines)
        return result
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        self._value = PyExpr(value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return self._value
    def get_value(self) -> PyExpr:
        "Return current option value"
        return self._value
    def set_value(self, value: PyExpr) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value
    value: PyExpr = property(get_value, set_value)

class PyCodeOption(Option[PyCode]):
    """String configuration option with Python code value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    required: True if option must have a value.
    default: Default option value.

Important:
    Python code must be properly indented, but ConfigParser multiline string values have
    leading whitespace removed. To circumvent this, the `PyCodeOption` supports assignment
    of text values where lines start with `|` character. This character is removed, along
    with any number of subsequent whitespace characters that are between `|` and first
    non-whitespace character on first line starting with `|`.
"""
    def __init__(self, name: str, description: str, *, required: bool=False, default: PyCode=None):
        self._value: PyCode = None
        super().__init__(name, PyCode, description, required, default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = self._value
        if '\n' in result:
            lines = []
            for line in result.splitlines(True):
                if lines:
                    lines.append('   | ' + line)
                else:
                    lines.append(line)
            result = ''.join(lines)
        return result
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        value = unindent_verticals(value)
        self._value = PyCode(value)
    def get_as_str(self) -> str:
        """Return value as string."""
        return self._value
    def get_value(self) -> PyCode:
        "Return current option value"
        return self._value
    def set_value(self, value: PyCode) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value
    value: PyCode = property(get_value, set_value)

class PyCallableOption(Option[PyCallable]):
    """String configuration option with Python callable value.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    signature: Callable signature.
    required: True if option must have a value.
    default: Default option value.

Attributes:
    signature (inspect.Signature): Callable signature.

Important:
    Python code must be properly indented, but `ConfigParser` multiline string values have
    leading whitespace removed. To circumvent this, the `PyCodeOption` supports assignment
    of text values where lines start with `|` character. This character is removed, along
    with any number of subsequent whitespace characters that are between `|` and first
    non-whitespace character on first line starting with `|`.
"""
    def __init__(self, name: str, description: str, signature: Signature, * ,
                 required: bool=False, default: PyCallable=None):
        self._value: PyCallable = None
        self.signature: Signature = signature
        super().__init__(name, PyCallable, description, required=required, default=default)
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = self._value
        if '\n' in result:
            lines = []
            for line in result.splitlines(True):
                if lines:
                    lines.append('   | ' + line)
                else:
                    lines.append(line)
            result = ''.join(lines)
        return result
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        value = unindent_verticals(value)
        self.set_value(PyCallable(value))
    def get_as_str(self) -> str:
        """Return value as string."""
        return self._value
    def get_value(self) -> PyCallable:
        "Return current option value"
        return self._value
    def set_value(self, value: PyCallable) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the callable has wrong signature.
"""
        self._check_value(value)
        if value is not None:
            val_sig = signature(value._callable_)
            if not _eq(val_sig.return_annotation, self.signature.return_annotation):
                raise ValueError("Wrong callable return type")
            if len(val_sig.parameters) != len(self.signature.parameters):
                raise ValueError("Wrong number of parameters")
            for par in self.signature.parameters.values():
                val_par: Parameter = val_sig.parameters[cast(Signature, par).name]
                if not _eq(val_par.annotation, cast(Signature, par).annotation):
                    raise ValueError(f"Wrong type, parameter '{val_par.name}'")
                if not _eq(val_par.default, cast(Signature, par).default):
                    raise ValueError(f"Wrong default, parameter '{val_par.name}'")
                if not _eq(val_par.kind, cast(Signature, par).kind):
                    raise ValueError(f"Wrong parameter kind, parameter '{val_par.name}'")
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value
    value: PyCallable = property(get_value, set_value)

class ConfigOption(Option[str]):
    """Configuration option with `Config` value.

Important:
    This option is intended for sub-configs that should have *configurable* name (i.e. the
    section name that holds sub-config values). To create sub-configs with fixed section
    names, simply assign them to instance attributes of `Config` instance that owns them
    (preferably in constructor).

    While the `value` attribute for this option is an instance of any class inherited from
    `Config`, in other ways it behaves like `StrOption` that loads/saves only name of its
    `Config` value (i.e. the section name). The actual I/O for sub-config's options is
    delegated to `Config` instance that owns this option.

    The "empty" value for this option is not `None` (because the `Config` instance always
    exists), but an empty string for `Config.name` attribute.

Arguments:
    name: Option name.
    description: Option description. Can span multiple lines.
    config: Option's value.
    required: True if option must have a value.
    default: Default `Config.name` value.
"""
    def __init__(self, name: str, description: str, config: Config, *, required: bool=False,
                 default: str=None):
        assert isinstance(config, Config)
        self._value: Config = config
        super().__init__(name, str, description, required, default)
    def validate(self) -> None:
        """Validates option state.

Raises:
    Error: When required option does not have a value.
"""
        if self.required and self.get_value().name == '':
            raise Error(f"Missing value for required option '{self.name}'")
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Note:
   This method calls `~Config.clear(to_default)`.

Arguments:
    to_default: If True, sets the `Config.name` to default value, else to empty string.
"""
        self._value.clear(to_default)
        self._value.name = self.default if to_default else ''
    def get_formatted(self) -> str:
        """Return value formatted for use in config file.

The string contains section name that will be used to store the `Config` values.
"""
        return self._value.name
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Important:
    Because the actual value is a `Config` instance, the string must contain the
    `Config.name` value (which is the section name used to store `Config` options).
    Beware that multiple Config instances with the same (section) name may cause
    collision when configuration is written to protobuf message or configuration file.

Arguments:
    value: New `Config.name` value.
"""
        self._value.name = value
    def get_as_str(self) -> str:
        """Return value as string.

Important:
    Because the actual value is a `Config` instance, the returned string is the section
    name used to store `Config` options.
"""
        return self._value.name
    def get_value(self) -> Config:
        "Return current option value."
        return self._value
    def set_value(self, value: str) -> None:
        """Set new option value.

This option type does not support direct assignment of `Config` value. Because this method
is also used to assign default value (which is a `Config.name`), it accepts None or string
argument that is interpreted as new Config name. `None` value is translated to empty string.

Arguments:
    value: New `Config` name.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When None or empty string is passed and option value is required.
"""
        if value is None:
            value = ''
        if value == '' and self.required:
            raise ValueError(f"Value is required for option '{self.name}'.")
        self._value.name = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            proto.options[self.name].as_string = self._value.name
    value: Config = property(get_value, set_value)

class ConfigListOption(Option[List]):
    """Configuration option with list of `Config` values.

Important:
    This option is intended for configurable set of sub-configs of fixed type.

    While the `value` attribute for this option is a list of instances of single class
    inherited from `Config`, in other ways it behaves like `ListOption` with `str` items
    that loads/saves only names of its `Config` items (i.e. the section names). The actual
    I/O for sub-config options is delegated to `Config` instance that owns this option.

Arguments:
    name:        Option name.
    description: Option description. Can span multiple lines.
    item_type:   Datatype of list items. Must be subclass of `Config`.
    required:    True if option must have a value.
    separator:   String that separates values when options value is read from `ConfigParser`.
                 It's possible to use a line break as separator.
                 If separator is `None` [default] and the value contains line breaks, it uses
                 the line break as separator, otherwise it uses comma as separator.

Attributes:
    item_type:   Datatype of list items.
    separator:   String that separates values when options value is read from `ConfigParser`.
                 Default separator is None. It's possible to use a line break as separator.
                 If separator is `None` and the value contains line breaks, it uses
                 the line break as separator, otherwise it uses comma as separator.

Important:
    When option is read from `ConfigParser`, empty values are ignored.
"""
    def __init__(self, name: str, description: str, item_type: Type[Config], *,
                 required: bool=False, separator: str=None):
        assert issubclass(item_type, Config)
        self._value: List = []
        self.item_type: Type[Config] = item_type
        self.separator: Optional[str] = separator
        super().__init__(name, list, description, required, [])
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = []
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = [i.name for i in self._value]
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        if sep == '\n':
            x = '\n   '
            return f"\n   {x.join(result)}"
        return f'{sep} '.join(result)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        new = []
        if value.strip():
            separator = ('\n' if '\n' in value else ',') if self.separator is None else self.separator
            for item in (i for i in value.split(separator) if i.strip()):
                new.append(self.item_type(item.strip()))
            self._value = new
    def get_as_str(self) -> str:
        """Return value as string."""
        result = [i.name for i in self._value]
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        return sep.join(result)
    def get_value(self) -> List:
        "Return current option value"
        return self._value
    def set_value(self, value: List) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        if value is not None:
            i = 0
            for item in value:
                if item.__class__ is not self.item_type:
                    raise ValueError(f"List item[{i}] has wrong type")
                i += 1
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        result = [i.name for i in self._value]
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        proto.options[self.name].as_string = sep.join(result)
    value: List = property(get_value, set_value)

class DataclassOption(Option[Any]):
    """Configuration option with a dataclass value.

The `ConfigParser` format for this option is a list of values, where each list items
defines value for dataclass field in `field_name:value_as_str` format. The configuration
must contain values for all fields for the dataclass that does not have default value.

Important:
    This option uses type annotation for dataclass to determine the actual data type for
    conversion from string. It means that:

    1. If type annotation contains "typing" types, it's necessary to specify "real" types
       for all dataclass fields using the `fields` argument.
    2. All used data types must have string convertors registered in `strconv` module.

Arguments:
    name:        Option name.
    dataclass:   Dataclass type.
    description: Option description. Can span multiple lines.
    required:    True if option must have a value.
    default:     Default option value.
    separator:   String that separates dataclass field values when options value is read
                 from `ConfigParser`. It's possible to use a line break as separator.
                 If separator is `None` [default] and the value contains line breaks, it
                 uses the line break as separator, otherwise it uses comma as separator.
    fields:      Dictionary that maps dataclass field names to data types.

Attributes:
    dataclass:   Dataclass type.
    separator:   String that separates dataclass field values when options value is read
                 from `ConfigParser`. Default separator is None. It's possible to use a line
                 break as separator. If separator is `None` and the value contains line
                 breaks, it uses the line break as separator, otherwise it uses comma as
                 separator.

Important:
    When option is read from `ConfigParser`, empty values are ignored.
"""
    def __init__(self, name: str, dataclass: Type, description: str, *, required: bool=False,
                 default: Any=None, separator: str=None, fields: Dict[str, Type]=None):
        assert hasattr(dataclass, '__dataclass_fields__')
        self._fields: Dict[str, Type] = get_type_hints(dataclass) if fields is None else fields
        if __debug__:
            for ftype in self._fields.values():
                assert get_convertor(ftype) is not None
        self._value: Any = None
        self.dataclass: Type = dataclass
        self.separator: Optional[str] = separator
        super().__init__(name, dataclass, description, required, default)
    def _get_str_fields(self) -> List[str]:
        result = []
        if self._value is not None:
            for fname in self._fields:
                result.append(f'{fname}:{convert_to_str(getattr(self._value, fname))}')
        return result
    def clear(self, to_default: bool=True) -> None:
        """Clears the option value.

Arguments:
    to_default: If True, sets the option value to default value, else to None.
"""
        self._value = self.default if to_default else None
    def get_formatted(self) -> str:
        """Return value formatted for use in config file."""
        if self._value is None:
            return '<UNDEFINED>'
        result = self._get_str_fields()
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        if sep == '\n':
            x = '\n   '
            return f"\n   {x.join(result)}"
        return f'{sep} '.join(result)
    def set_as_str(self, value: str) -> None:
        """Set new option value from string.

Arguments:
    value: New option value.

Raises:
    ValueError: When the argument is not a valid option value.
"""
        new = {}
        if value.strip():
            separator = ('\n' if '\n' in value else ',') if self.separator is None else self.separator
            for item in (i for i in value.split(separator) if i.strip()):
                try:
                    field_name, field_value = item.split(':', 1)
                except Exception:
                    raise ValueError(f"Illegal value '{value}' for option '{self.name}'")
                field_name = field_name.strip()
                ftype = self._fields.get(field_name)
                if ftype is None:
                    raise ValueError(f"Unknown data field '{field_name}' for option '{self.name}'")
                convertor = get_convertor(ftype)
                new[field_name] = convertor.from_str(ftype, field_value.strip())
                try:
                    new_val = self.dataclass(**new)
                except Exception:
                    raise ValueError(f"Illegal value '{value}' for option '{self.name}'")
            self._value = new_val
    def get_as_str(self) -> str:
        """Return value as string."""
        result = self._get_str_fields()
        sep = self.separator
        if sep is None:
            sep = '\n' if sum(len(i) for i in result) > 80 else ','
        return sep.join(result)
    def get_value(self) -> Any:
        "Return current option value"
        return self._value
    def set_value(self, value: Any) -> None:
        """Set new option value.

Arguments:
    value: New option value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        self._check_value(value)
        self._value = value
    def load_proto(self, proto: ConfigProto) -> None:
        """Deserialize value from `.ConfigProto` message.

Arguments:
    proto: Protobuf message that may contains options value.

Raises:
    TypeError: When the new value is of the wrong type.
    ValueError: When the argument is not a valid option value.
"""
        if self.name in proto.options:
            opt = proto.options[self.name]
            if opt.HasField('as_string'):
                self.set_as_str(opt.as_string)
            else:
                raise TypeError(f"Wrong value type: {opt.WhichOneof('kind')[3:]}")
    def save_proto(self, proto: ConfigProto) -> None:
        """Serialize value into `.ConfigProto` message.

Arguments:
    proto: Protobuf message where option value should be stored.
"""
        if self._value is not None:
            result = self._get_str_fields()
            sep = self.separator
            if sep is None:
                sep = '\n' if sum(len(i) for i in result) > 80 else ','
            proto.options[self.name].as_string = sep.join(result)
    value: Any = property(get_value, set_value)




#coding:utf-8
"""A setuptools based setup module for firebird-core package.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup

setup()
